#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <syslog.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/poll.h>
#include <sys/mman.h>
#include <sys/file.h>
#include <sys/time.h>
#include <time.h>

#define LOGFILE_PATH		"/tmp/logfile"	/* FIXME */

#define LOG_START		8
#define _1MB_			(1024 * 1024)
#define DEFAULT_LOGSIZE		(5 * _1MB_)

static int
mutex_enter (int logfd)
{
	int err = 0;

	if (-1 == lockf(logfd, F_LOCK, 0)) {
		err = errno;
		perror("flock()");
	}
	return err;
}

static int
mutex_exit (int logfd)
{
	int err = 0;

	if (-1 == lockf(logfd, F_ULOCK, 0)) {
		err = errno;
		perror("flock()");
	}
	return err;
}

static int
get_logbuf_offset (char *logbuf, int logsize)
{
	int offset = atoi(logbuf);
	if (offset < 0 || (LOG_START + offset) >= logsize)
		offset = 0;
	return offset;
}

static int
write_log_byte (char *logbuf, int logsize, int offset, char c)
{
	logbuf[LOG_START + offset] = c;
	offset += 1;
	if ((LOG_START + offset) >= logsize)
		offset = 0;
	return offset;
}

static void
write_log_string (char *logbuf, int logsize, const char *msg)
{
	int offset;

	offset = get_logbuf_offset(logbuf, logsize);
	while (*msg)
		offset = write_log_byte(logbuf, logsize, offset, *msg++);
	snprintf(logbuf, LOG_START, "%07u", offset);
	logbuf[LOG_START-1] = '\n';
}

static void
write_timestamp (char *logbuf, int logsize)
{
	time_t		now;
	struct tm	lt;
	char		ts[32];

	now = time(NULL);
	localtime_r(&now, &lt);
	strftime(ts, sizeof(ts), "%F %T: ", &lt);
	write_log_string(logbuf, logsize, ts);
}

static void
write_logtext (char *logbuf, int logsize, const char *logtext)
{
	write_timestamp (logbuf, logsize);
	write_log_string(logbuf, logsize, logtext);
	write_log_string(logbuf, logsize, "\n");
}

static char *
path_tail (char *path)
{
	char *p;

	// return pointer to the final component of a pathname
	p = strrchr(path, '/');
	if (p && *++p)
		return p;
	return path;
}

static void
dump_logfile (char *logbuf, int logsize)
{
	int	found_first_newline;
	char	last_c = 0;
	int	offset = get_logbuf_offset(logbuf, logsize);
	int	start  = offset;

	found_first_newline = (logbuf[logsize-1] == '\0');

	do {
		unsigned char c = logbuf[LOG_START + offset];
		if (found_first_newline) {
			if ((c >= ' ' && c <= 0x7f) || c == '\n') {
				last_c = c;
				putchar(c);
			}
		}
		if (c == '\n')
			found_first_newline = 1;
		offset += 1;
		if ((LOG_START + offset) >= logsize)
			offset = 0;
	} while (offset != start);

	if (last_c != '\n')
		putchar('\n');
}

int
main (int argc, char *argv[])
{
	int		err, err2;
	const char	*logtext, logfile[] = LOGFILE_PATH;
	char		*logbuf;
	int		logfd, logsize;
	struct stat	st;

	if (argc != 2) {
		const char *myname = path_tail(argv[0]);
		fprintf(stderr, "Usage:  %s  [--dump | \"log text\"]\n", myname);
		exit(EINVAL);
	}
	logtext = argv[1];

	logfd = open(logfile, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR);
	if (logfd == -1) {
		int err = errno;
		perror(logfile);
		exit(err);
	}

	err = mutex_enter(logfd);
	if (err)
		exit(err);

	do {
		if (-1 == fstat(logfd, &st)) {
			int err = errno;
			perror(logfile);
			mutex_exit(logfd);
			exit(err);
		}
		logsize = st.st_size;
		if (logsize == 0) {
			if (-1 == ftruncate(logfd, DEFAULT_LOGSIZE)) {
				int err = errno;
				perror(logfile);
				mutex_exit(logfd);
				exit(err);
			}
		}
	} while (logsize == 0);

	logbuf = mmap(NULL, logsize, PROT_READ|PROT_WRITE, MAP_SHARED, logfd, 0);
	if (logbuf == MAP_FAILED) {
		err = errno;
		perror("mmap(logfile)");
	} else {
	       	if (0 == strcmp(logtext, "--dump"))
			dump_logfile(logbuf, logsize);
		else
			write_logtext(logbuf, logsize, logtext);
		if (-1 == munmap(logbuf, logsize)) {
			err = errno;
			perror("munmap(logfile)");
		}
	}
	err2 = mutex_exit(logfd);
	if (!err)
		err = err2;
	if (-1 == close(logfd)) {
		if (!err)
			err = errno;
		perror("close(logfile)");
	}
	exit(err);
}

